# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from collections import OrderedDict
from contextlib import contextmanager
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

from deprecation import deprecated
from pydantic import validate_arguments

from kolena._api.v1.fr import TestSuite as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.endpoints import get_test_suite_url
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import ValidatorConfig
from kolena.errors import NotFoundError
from kolena.fr.test_case import TestCase


class TestSuite(ABC, Frozen, WithTelemetry):
    """
    A test suite groups together one or more test cases.
    """

    name: str
    """The unique name of this test suite."""

    version: int
    """
    The version of this test suite. A test suite's version is automatically incremented whenever it is edited via
    [`TestSuite.edit`][kolena.fr.TestSuite.edit].
    """

    description: str
    """
    Free-form, human-readable description of this test suite. Can be edited at any time via
    [`TestSuite.edit`][kolena.fr.TestSuite.edit].
    """

    baseline_test_cases: List[TestCase]
    """The baseline [`TestCase`][kolena.fr.TestCase] object(s) for this test suite."""

    non_baseline_test_cases: List[TestCase]
    """The non-baseline [`TestCase`][kolena.fr.TestCase] object(s) for this test suite."""

    baseline_image_count: int
    """The number of images attached to the baseline test case(s)."""

    baseline_pair_count_genuine: int
    """The number of genuine pairs attached to the baseline test case(s)."""

    baseline_pair_count_imposter: int
    """The count of imposter pairs attached to the baseline test case(s)."""

    Data = API.EntityData

    _id: int
    _data: API.EntityData

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        baseline_test_cases: Optional[List[TestCase]] = None,
        non_baseline_test_cases: Optional[List[TestCase]] = None,
        reset: bool = False,
    ):
        try:
            self._populate_from_other(self.load(name, version))
            if description is not None and self.description != description and not reset:
                log.warn("test suite already exists, not updating description when reset=False")
            if baseline_test_cases is not None or non_baseline_test_cases is not None:
                if self.version > 0 and not reset:
                    log.warn("test suite already exists, not updating test cases when reset=False")
                else:
                    self._hydrate(baseline_test_cases, non_baseline_test_cases, description)
        except NotFoundError:
            if version is not None:
                log.warn(f"creating new test suite '{name}', ignoring provided version")
            self._populate_from_other(self.create(name, description, baseline_test_cases, non_baseline_test_cases))
        self._freeze()

    @property
    @deprecated(details="use `TestSuite` instance attributes", deprecated_in="0.57.0")
    def data(self) -> API.EntityData:
        """
        !!! warning "Deprecated: since `0.57.0`"
            Access this data via instance attributes, e.g. `self.baseline_test_cases`, directly.

        The data associated with this test suite.
        """
        return self._data

    @data.setter
    @deprecated(details="use `TestSuite` instance attributes", deprecated_in="0.57.0")
    def data(self, new_data: API.EntityData) -> None:
        self._data = new_data

    @classmethod
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        baseline_test_cases: Optional[List[TestCase]] = None,
        non_baseline_test_cases: Optional[List[TestCase]] = None,
    ) -> "TestSuite":
        """
        Create a new test suite with the provided name.

        :param name: The name of the new test suite to create.
        :param description: Optional free-form description of the test suite to create.
        :param baseline_test_cases: Optionally specify a list of test cases to use as baseline for the test suite.
        :param non_baseline_test_cases: Optionally specify a list of test cases to populate the test suite.
        :return: The newly created test suite.
        """
        request = API.CreateRequest(name=name, description=description or "")
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=API.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        obj._hydrate(baseline_test_cases, non_baseline_test_cases)
        return obj

    @classmethod
    def load(cls, name: str, version: Optional[int] = None) -> "TestSuite":
        """
        Load an existing test suite with the provided name.

        :param name: the name of the test suite to load.
        :param version: optionally specify a particular version of the test suite to load. Defaults to the latest
            version when unset.
        :return: the loaded test suite.
        """
        return cls._load_by_name(name, version)

    @classmethod
    @deprecated(details="use `TestSuite.load`", deprecated_in="0.57.0")
    def load_by_name(cls, name: str, version: Optional[int] = None) -> "TestSuite":
        """
        !!! warning "Deprecated: since `0.57.0`"
            Use [`TestSuite.load`][kolena.fr.TestSuite.load] instead.

        Retrieve the existing test suite with the provided name.

        :param name: Name of the test suite to retrieve.
        :param version: Optionally specify the version of the named test suite to retrieve. When absent the latest
            version of the test suite is returned.
        :return: The retrieved test suite.
        :raises NotFoundError: If the test suite with the provided name doesn't exist.
        """
        return cls.load(name, version)

    @classmethod
    def _load_by_name(cls, name: str, version: Optional[int] = None) -> "TestSuite":
        request = API.LoadByNameRequest(name=name, version=version)
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = cls._create_from_data(from_dict(data_class=API.EntityData, data=res.json()))
        log.info(f"loaded test suite '{name}' (v{obj.version}) ({get_test_suite_url(obj._id)})")
        return obj

    def _populate_from_other(self, other: "TestSuite") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self.baseline_test_cases = other.baseline_test_cases
            self.non_baseline_test_cases = other.non_baseline_test_cases
            self.baseline_image_count = other.baseline_image_count
            self.baseline_pair_count_genuine = other.baseline_pair_count_genuine
            self.baseline_pair_count_imposter = other.baseline_pair_count_imposter
            self.data = other._data

    @classmethod
    def _create_from_data(cls, data: API.EntityData) -> "TestSuite":
        obj = cls.__new__(cls)
        obj._id = data.id
        obj.name = data.name
        obj.version = data.version
        obj.description = data.description
        obj.baseline_test_cases = [TestCase._create_from_data(tc) for tc in data.baseline_test_cases]
        obj.non_baseline_test_cases = [TestCase._create_from_data(tc) for tc in data.non_baseline_test_cases]
        obj.baseline_image_count = data.baseline_image_count
        obj.baseline_pair_count_genuine = data.baseline_pair_count_genuine
        obj.baseline_pair_count_imposter = data.baseline_pair_count_imposter
        obj.data = data
        obj._freeze()
        return obj

    def _hydrate(
        self,
        baseline_test_cases: Optional[List[TestCase]] = None,
        non_baseline_test_cases: Optional[List[TestCase]] = None,
        description: Optional[str] = None,
    ) -> None:
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            if baseline_test_cases is not None:
                for test_case in baseline_test_cases:
                    editor.add(test_case, True)
            if non_baseline_test_cases is not None:
                for test_case in non_baseline_test_cases:
                    editor.add(test_case, False)

    class Editor:
        """Interface to edit a test suite. Create with [`TestSuite.edit`][kolena.fr.TestSuite.edit]."""

        _edited: bool
        _reset: bool
        _description: str
        _initial_description: str
        _baseline_test_cases: Dict[str, int]
        _non_baseline_test_cases: Dict[str, int]

        @validate_arguments(config=ValidatorConfig)
        def __init__(self, description: str, reset: bool) -> None:
            self._baseline_test_cases: Dict[str, int] = OrderedDict()  # map from name -> id
            self._non_baseline_test_cases: Dict[str, int] = OrderedDict()  # map from name -> id
            self._reset = reset
            self._description = description
            self._initial_description = description
            self._edited = False

        @validate_arguments(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """
            Update the description of the test suite.

            :param description: The new description of the test suite.
            """
            self._description = description
            self._edited = True

        @validate_arguments(config=ValidatorConfig)
        def add(self, test_case: TestCase, is_baseline: Optional[bool] = None) -> None:
            """
            Add a test case to this test suite. If a different version of the test case already exists in this test
            suite, it is replaced and its baseline status will be propagated when `is_baseline` is unset.

            :param test_case: The test case to add to the test suite.
            :param is_baseline: Specify that this test case is a part of the "baseline," i.e. if the samples in this
                test case should contribute to the computation of thresholds within this test suite.
            """
            name = test_case.name
            # clean up any previous versions and propagates its baseline status
            set_is_baseline: Optional[bool] = None
            if name in self._baseline_test_cases.keys():
                self._baseline_test_cases.pop(name)
                set_is_baseline = is_baseline if is_baseline is not None else True
            if name in self._non_baseline_test_cases.keys():
                self._non_baseline_test_cases.pop(name)
                set_is_baseline = is_baseline if is_baseline is not None else False

            is_baseline_ret = set_is_baseline or is_baseline or False
            self._add(test_case, is_baseline=is_baseline_ret)

        @validate_arguments(config=ValidatorConfig)
        def remove(self, test_case: TestCase) -> None:
            """
            Remove the provided [`TestCase`][kolena.fr.TestCase] from the test suite. Any version of this test case in
            the suite will be removed; the version does not need to match exactly.

            :param test_case: The test case to remove.
            """
            name = test_case.name
            if name in self._baseline_test_cases.keys():
                self._baseline_test_cases.pop(name)
            elif name in self._non_baseline_test_cases.keys():
                self._non_baseline_test_cases.pop(name)
            else:
                raise KeyError(f"test case '{name}' not in test suite")
            self._edited = True

        @deprecated(details="use `TestSuite.Editor.add`", deprecated_in="0.57.0")
        @validate_arguments(config=ValidatorConfig)
        def merge(self, test_case: TestCase, is_baseline: Optional[bool] = None) -> None:
            """
            !!! warning "Deprecated: since `0.57.0`"
                Replaced by idempotent behavior in [`TestSuite.Editor.add`][kolena.fr.TestSuite.Editor.add].

            Add the [`TestCase`][kolena.fr.TestCase] to the suite. If a test case by this name already exists in the
            suite, replace the previous version of that test case with the newly provided version.

            :param test_case: The test case to be merged into the test suite.
            :param is_baseline: Optionally specify whether or not this test case should be considered as a part of the
                baseline for this test suite. When not specified, the previous value for `is_baseline` for this test
                case in this test suite is propagated forward. Defaults to false if the test case does not already exist
                in this suite.
            """
            self.add(test_case, is_baseline)

        @validate_arguments(config=ValidatorConfig)
        def _add(self, test_case: TestCase, is_baseline: bool = False) -> None:
            name = test_case.name
            if is_baseline:
                self._baseline_test_cases[name] = test_case._id
            else:
                self._non_baseline_test_cases[name] = test_case._id
            self._edited = True

    @contextmanager
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Context-managed way to perform many modification options on a test suite and commit the results when the context
        is exited, resulting in a single version bump.

        ```python
        with TestSuite.load("my-test-suite").edit() as editor:
            editor.add(TestCase.load("my-test-case"))
        ```
        """
        editor = TestSuite.Editor(self.description, reset)
        if not reset:
            for baseline_test_case in self.baseline_test_cases:
                editor.add(baseline_test_case, is_baseline=True)
            for non_baseline_test_case in self.non_baseline_test_cases:
                editor.add(non_baseline_test_case, is_baseline=False)
            editor._edited = False

        yield editor

        # no-op contexts have no effect, do not bump version
        if not editor._edited:
            return

        log.info(f"editing test suite '{self.name}' (v{self.version})")
        request = API.EditRequest(
            test_suite_id=self._id,
            current_version=self.version,
            name=self.name,
            description=editor._description,
            baseline_test_case_ids=list(editor._baseline_test_cases.values()),
            non_baseline_test_case_ids=list(editor._non_baseline_test_cases.values()),
        )
        res = krequests.post(endpoint_path=API.Path.EDIT.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        test_suite_data = from_dict(data_class=API.EntityData, data=res.json())
        self._populate_from_other(self._create_from_data(test_suite_data))
        log.success(f"edited test suite '{self.name}' (v{self.version}) ({get_test_suite_url(self._id)})")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import cast
from typing import Type

import pandas as pd
import pandera as pa
from pandera.typing import Series

from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.datatypes import LoadableDataFrame

JSONObject = object


class LoadTestImagesDataFrameSchema(pa.SchemaModel):
    test_sample_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    locator: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    dataset: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)
    metadata: Series[JSONObject] = pa.Field(coerce=True)


class LoadTestImagesDataFrame(
    pa.typing.DataFrame[LoadTestImagesDataFrameSchema],
    LoadableDataFrame["LoadTestImagesDataFrame"],
):
    @classmethod
    def get_schema(cls) -> Type[LoadTestImagesDataFrameSchema]:
        return LoadTestImagesDataFrameSchema

    @classmethod
    def from_serializable(cls, df_samples: pd.DataFrame) -> "LoadTestImagesDataFrame":
        df = df_samples.copy()
        df["locator"] = df["data"].apply(lambda v: json.loads(v)["locator"])
        df["metadata"] = df["metadata"].apply(json.loads)
        df = df[["test_sample_id", "locator", "dataset", "metadata"]]
        return cast(LoadTestImagesDataFrame, validate_df_schema(df, LoadTestImagesDataFrameSchema, trusted=True))



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass


@dataclass(frozen=True)
class ValidateRequest:
    api_token: str
    version: str


@dataclass(frozen=True)
class ValidateResponse:
    tenant: str
    access_token: str
    token_type: str
    tenant_telemetry: bool



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
from typing import Any
from typing import Generic
from typing import Type
from typing import TypeVar

from kolena.errors import DirectInstantiationError
from kolena.errors import IncorrectUsageError

T = TypeVar("T")
U = TypeVar("U", bound="Uninstantiable")


class Uninstantiable(Generic[T]):
    __slots__ = ("data",)
    __frozen__ = False

    data: T

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise DirectInstantiationError(
            f"{type(self).__name__} is not directly instantiatable, "
            "please use one of the available static constructors",
        )

    @typing.no_type_check
    def __setattr__(self, key, value):
        if self.__frozen__:
            raise IncorrectUsageError("cannot modify frozen class")
        object.__setattr__(self, key, value)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(data={repr(self.data)})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and other.data == self.data

    @classmethod
    def __factory__(cls: Type[U], data: Any) -> U:
        obj = cls.__new__(cls)
        obj.data = data
        obj.__frozen__ = True
        return obj

    def __update__(self, data: T) -> None:
        object.__setattr__(self, "__frozen__", False)
        self.data = data
        self.__frozen__ = True



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np

from kolena._experimental.object_detection.utils import compute_average_precision
from kolena._experimental.object_detection.utils import compute_confusion_matrix_plot
from kolena._experimental.object_detection.utils import compute_f1_plot
from kolena._experimental.object_detection.utils import compute_f1_plot_multiclass
from kolena._experimental.object_detection.utils import compute_optimal_f1_threshold_multiclass
from kolena._experimental.object_detection.utils import compute_pr_curve
from kolena._experimental.object_detection.utils import compute_pr_plot
from kolena._experimental.object_detection.utils import compute_pr_plot_multiclass
from kolena._experimental.object_detection.workflow import ClassMetricsPerTestCase
from kolena._experimental.object_detection.workflow import GroundTruth
from kolena._experimental.object_detection.workflow import Inference
from kolena._experimental.object_detection.workflow import TestCase
from kolena._experimental.object_detection.workflow import TestCaseMetrics
from kolena._experimental.object_detection.workflow import TestSample
from kolena._experimental.object_detection.workflow import TestSampleMetrics
from kolena._experimental.object_detection.workflow import TestSuite
from kolena._experimental.object_detection.workflow import TestSuiteMetrics
from kolena._experimental.object_detection.workflow import ThresholdConfiguration
from kolena._experimental.object_detection.workflow import ThresholdStrategy
from kolena.workflow import Evaluator
from kolena.workflow import Plot
from kolena.workflow.annotation import ScoredClassificationLabel
from kolena.workflow.metrics import f1_score as compute_f1_score
from kolena.workflow.metrics import match_inferences_multiclass
from kolena.workflow.metrics import MulticlassInferenceMatches
from kolena.workflow.metrics import precision as compute_precision
from kolena.workflow.metrics import recall as compute_recall


class ObjectDetectionEvaluator(Evaluator):
    """
    This `ObjectDetectionEvaluator` transforms inferences into metrics for the object detection workflow for a single
    class or multiple classes.

    When a [`ThresholdConfiguration`][kolena._experimental.object_detection.workflow.ThresholdConfiguration] is
    configured to use an F1-Optimal threshold strategy, the evaluator requires that the first test case retrieved for
    a test suite contains the complete sample set.

    For additional functionality, see the associated [base class documentation][kolena.workflow.evaluator.Evaluator].
    """

    # Assumes that the first test case retrieved for the test suite contains the complete sample set to be used for
    # F1-Optimal threshold computation. Subsequent requests for a given threshold strategy (for other test cases) will
    # hit this cache and use the previously computed population level confidence thresholds.
    threshold_cache: Dict[str, Dict[str, float]] = {}  # configuration -> label -> threshold

    # Keeps track of test sample locators for each test case (used for total # of image count in aggregated metrics)
    locators_by_test_case: Dict[str, List[str]] = {}

    def compute_image_metrics(
        self,
        ground_truth: GroundTruth,
        inference: Inference,
        configuration: ThresholdConfiguration,
        labels: Set[str],  # the labels being tested in this test case
    ) -> TestSampleMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        bbox_matches: MulticlassInferenceMatches = match_inferences_multiclass(
            ground_truth.bboxes,
            [
                inf
                for inf in inference.bboxes
                if inf.label in labels and inf.score >= configuration.min_confidence_score
            ],
            ignored_ground_truths=ground_truth.ignored_bboxes,
            mode="pascal",
            iou_threshold=configuration.iou_threshold,
        )
        image_labels = {gt.label for gt in ground_truth.bboxes}
        tp = [inf for _, inf in bbox_matches.matched if inf.score >= thresholds[inf.label]]
        fp = [
            inf
            for inf in bbox_matches.unmatched_inf
            if inf.label in image_labels and inf.score >= thresholds[inf.label]
        ]
        fn = [gt for gt, _ in bbox_matches.unmatched_gt] + [
            gt for gt, inf in bbox_matches.matched if inf.score < thresholds[inf.label]
        ]
        confused = [
            inf for _, inf in bbox_matches.unmatched_gt if inf is not None and inf.score >= thresholds[inf.label]
        ]
        non_ignored_inferences = tp + fp
        scores = [inf.score for inf in non_ignored_inferences]

        fields = [
            ScoredClassificationLabel(label=label, score=thresholds[label])
            for label in thresholds.keys()
            if label in image_labels
        ]
        return TestSampleMetrics(
            TP=tp,
            FP=fp,
            FN=fn,
            Confused=confused,
            count_TP=len(tp),
            count_FP=len(fp),
            count_FN=len(fn),
            count_Confused=len(confused),
            has_TP=len(tp) > 0,
            has_FP=len(fp) > 0,
            has_FN=len(fn) > 0,
            has_Confused=len(confused) > 0,
            max_confidence_above_t=max(scores) if len(scores) > 0 else None,
            min_confidence_above_t=min(scores) if len(scores) > 0 else None,
            thresholds=fields,
        )

    def compute_and_cache_f1_optimal_thresholds(
        self,
        configuration: ThresholdConfiguration,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
    ) -> None:
        if configuration.threshold_strategy != ThresholdStrategy.F1_OPTIMAL:
            return

        if configuration.display_name() in self.threshold_cache.keys():
            return

        all_bbox_matches = [
            match_inferences_multiclass(
                ground_truth.bboxes,
                [inf for inf in inference.bboxes if inf.score >= configuration.min_confidence_score],
                ignored_ground_truths=ground_truth.ignored_bboxes,
                mode="pascal",
                iou_threshold=configuration.iou_threshold,
            )
            for _, ground_truth, inference in inferences
        ]
        optimal_thresholds = compute_optimal_f1_threshold_multiclass(all_bbox_matches)
        self.threshold_cache[configuration.display_name()] = optimal_thresholds

    def compute_test_sample_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> List[Tuple[TestSample, TestSampleMetrics]]:
        assert configuration is not None, "must specify configuration"
        # compute thresholds to cache values for subsequent steps
        self.compute_and_cache_f1_optimal_thresholds(configuration, inferences)
        labels = {gt.label for _, gts, _ in inferences for gt in gts.bboxes}
        return [(ts, self.compute_image_metrics(gt, inf, configuration, labels)) for ts, gt, inf in inferences]

    def compute_aggregate_label_metrics(
        self,
        matchings: List[MulticlassInferenceMatches],
        label: str,
        thresholds: Dict[str, float],
    ) -> ClassMetricsPerTestCase:
        match_matched = []
        match_unmatched_gt = []
        match_unmatched_inf = []
        samples_with_label = 0
        # filter the matching to only consider one class
        for match in matchings:
            sample_flag = False
            for gt, inf in match.matched:
                if gt.label == label:
                    sample_flag = True
                    match_matched.append((gt, inf))
            for gt, inf in match.unmatched_gt:
                if gt.label == label:
                    sample_flag = True
                    match_unmatched_gt.append((gt, inf))
            for inf in match.unmatched_inf:
                if inf.label == label:
                    sample_flag = True
                    match_unmatched_inf.append(inf)
            if sample_flag:
                samples_with_label += 1

        all_bbox_matches_for_one_label = [
            MulticlassInferenceMatches(
                matched=match_matched,
                unmatched_gt=match_unmatched_gt,
                unmatched_inf=match_unmatched_inf,
            ),
        ]

        tp = [inf for _, inf in match_matched if inf.score >= thresholds[inf.label]]
        fp = [inf for inf in match_unmatched_inf if inf.score >= thresholds[inf.label]]
        fn = [gt for gt, _ in match_unmatched_gt] + [
            gt for gt, inf in match_matched if inf.score < thresholds[inf.label]
        ]
        confused = [inf for _, inf in match_unmatched_gt if inf is not None and inf.score >= thresholds[inf.label]]
        tp_count = len(tp)
        fp_count = len(fp)
        fn_count = len(fn)
        confused_count = len(confused)
        precision = compute_precision(tp_count, fp_count)
        recall = compute_recall(tp_count, fn_count)
        f1_score = compute_f1_score(tp_count, fp_count, fn_count)

        average_precision = 0.0
        if precision > 0:
            baseline_pr_curve = compute_pr_curve(all_bbox_matches_for_one_label)
            if baseline_pr_curve is not None:
                average_precision = compute_average_precision(baseline_pr_curve.y, baseline_pr_curve.x)

        return ClassMetricsPerTestCase(
            Class=label,
            nImages=samples_with_label,
            Threshold=thresholds[label],
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            Confused=confused_count,
            Precision=precision,
            Recall=recall,
            F1=f1_score,
            AP=average_precision,
        )

    def compute_test_case_metrics(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[TestSampleMetrics],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> TestCaseMetrics:
        assert configuration is not None, "must specify configuration"
        thresholds = self.get_confidence_thresholds(configuration)
        labels = {gt.label for _, gts, _ in inferences for gt in gts.bboxes}
        all_bbox_matches = [
            match_inferences_multiclass(
                ground_truth.bboxes,
                [
                    inf
                    for inf in inference.bboxes
                    if inf.label in labels and inf.score >= configuration.min_confidence_score
                ],
                ignored_ground_truths=ground_truth.ignored_bboxes,
                mode="pascal",
                iou_threshold=configuration.iou_threshold,
            )
            for _, ground_truth, inference in inferences
        ]

        # compute nested metrics per class
        per_class_metrics: List[ClassMetricsPerTestCase] = []
        for label in labels:
            metrics_per_class = self.compute_aggregate_label_metrics(all_bbox_matches, label, thresholds)
            per_class_metrics.append(metrics_per_class)

        self.locators_by_test_case[test_case.name] = [ts.locator for ts, _, _ in inferences]
        tp_count = sum(im.count_TP for im in metrics)
        fp_count = sum(im.count_FP for im in metrics)
        fn_count = sum(im.count_FN for im in metrics)

        return TestCaseMetrics(
            PerClass=per_class_metrics,
            Objects=tp_count + fn_count,
            Inferences=tp_count + fp_count,
            TP=tp_count,
            FN=fn_count,
            FP=fp_count,
            macro_Precision=np.mean([class_metrics.Precision for class_metrics in per_class_metrics]),
            macro_Recall=np.mean([class_metrics.Recall for class_metrics in per_class_metrics]),
            macro_F1=np.mean([class_metrics.F1 for class_metrics in per_class_metrics]),
            mean_AP=np.mean([class_metrics.AP for class_metrics in per_class_metrics]),
        )

    def compute_test_case_plots(
        self,
        test_case: TestCase,
        inferences: List[Tuple[TestSample, GroundTruth, Inference]],
        metrics: List[TestSampleMetrics],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> Optional[List[Plot]]:
        assert configuration is not None, "must specify configuration"
        labels = {gt.label for _, gts, _ in inferences for gt in gts.bboxes}
        thresholds = self.get_confidence_thresholds(configuration)
        all_bbox_matches = [
            match_inferences_multiclass(
                ground_truth.bboxes,
                [
                    inf
                    for inf in inference.bboxes
                    if inf.label in labels and inf.score >= configuration.min_confidence_score
                ],
                ignored_ground_truths=ground_truth.ignored_bboxes,
                mode="pascal",
                iou_threshold=configuration.iou_threshold,
            )
            for _, ground_truth, inference in inferences
        ]

        # clean matching for confusion matrix
        match_matched = [
            (gt, inf) for match in all_bbox_matches for gt, inf in match.matched if inf.score >= thresholds[inf.label]
        ]
        match_unmatched_gt = [
            (gt, inf)
            for match in all_bbox_matches
            for gt, inf in match.unmatched_gt
            if inf is not None and inf.score >= thresholds[inf.label]
        ]
        confusion_matrix_matchings = [
            MulticlassInferenceMatches(
                matched=match_matched,
                unmatched_gt=match_unmatched_gt,
                unmatched_inf=[],
            ),
        ]

        plots: Optional[List[Plot]] = []
        plots.extend(
            filter(
                None,
                [
                    compute_pr_plot(all_bbox_matches),
                    compute_f1_plot(all_bbox_matches),
                    compute_f1_plot_multiclass(all_bbox_matches),
                    compute_pr_plot_multiclass(all_bbox_matches),
                    compute_confusion_matrix_plot(confusion_matrix_matchings),
                ],
            ),
        )

        return plots

    def compute_test_suite_metrics(
        self,
        test_suite: TestSuite,
        metrics: List[Tuple[TestCase, TestCaseMetrics]],
        configuration: Optional[ThresholdConfiguration] = None,
    ) -> TestSuiteMetrics:
        assert configuration is not None, "must specify configuration"
        return TestSuiteMetrics(
            n_images=len({locator for tc, _ in metrics for locator in self.locators_by_test_case[tc.name]}),
            mean_AP=np.average([tcm.mean_AP for _, tcm in metrics]),
        )

    def get_confidence_thresholds(self, configuration: ThresholdConfiguration) -> Dict[str, float]:
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_03:
            return defaultdict(lambda: 0.3)
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_05:
            return defaultdict(lambda: 0.5)
        if configuration.threshold_strategy == ThresholdStrategy.FIXED_075:
            return defaultdict(lambda: 0.75)
        if configuration.threshold_strategy == ThresholdStrategy.F1_OPTIMAL:
            return self.threshold_cache[configuration.display_name()]
        raise RuntimeError(f"unrecognized threshold strategy: {configuration.threshold_strategy}")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import numpy as np
import pandas as pd
from pydantic import validate_arguments
from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad as LoadAPI
from kolena._api.v1.fr import TestImages as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.uninstantiable import Uninstantiable
from kolena._utils.validators import ValidatorConfig
from kolena.fr import TestCase
from kolena.fr import TestSuite
from kolena.fr.datatypes import TEST_IMAGE_COLUMNS
from kolena.fr.datatypes import TestImageDataFrame
from kolena.fr.datatypes import TestImageDataFrameSchema
from kolena.fr.datatypes import TestImageRecord


class TestImages(Uninstantiable[None]):
    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def load(
        cls,
        data_source: Optional[Union[str, TestSuite, TestSuite.Data, TestCase, TestCase.Data]] = None,
        include_augmented: bool = False,
    ) -> TestImageDataFrame:
        """
        Load a DataFrame describing images registered in the Kolena platform.

        :param data_source: Optionally specify the single data source to be retrieved, e.g. `"my-data-source"`.
            Alternatively, provide a [`TestSuite`][kolena.fr.TestSuite] or [`TestCase`][kolena.fr.TestCase] as source.
            If no argument is provided, all images registered using
            [`TestImages.register][kolena.fr.TestImages.register] are returned.
        :param include_augmented: Optionally specify that augmented images should be returned. By default, only
            original images are returned. Ignored when test case or test suite is provided as `data_source`.
        """
        log.info("loading test images")
        return _BatchedLoader.concat(
            cls.iter(data_source=data_source, include_augmented=include_augmented),
            TestImageDataFrame,
        )

    @dataclass(config=ValidatorConfig)
    class _Registrar:
        records: List[TestImageRecord]
        locators: Set[str]

    class Registrar(Uninstantiable[_Registrar]):
        @validate_arguments(config=ValidatorConfig)
        def add(
            self,
            locator: str,
            data_source: str,
            width: int,
            height: int,
            bounding_box: Optional[np.ndarray] = None,
            landmarks: Optional[np.ndarray] = None,
            tags: Optional[Dict[str, str]] = None,
        ) -> None:
            """
            Add a new image to Kolena. If the provided locator is already registered with the platform, its metadata
            will be updated.

            :param locator: Bucket locator for the provided image, e.g. `s3://bucket-name/path/to/image.jpg`.
            :param data_source: Name of the source for the image being registered.
            :param width: Width in pixels of the image being registered.
            :param height: Height in pixels of the image being registered.
            :param bounding_box: Optional 4-element array specifying the ground truth bounding box for this image, of
                the form `[top_left_x, top_left_y, bottom_right_x, bottom_right_y]`.
            :param landmarks: Optional 10-element array specifying (x, y) coordinates for five facial landmarks of the
                form `[left_eye_{x,y}, right_eye_{x,y}, nose_{x,y}, left_mouth_{x,y}, right_mouth_{x,y}]`.
            :param tags: Tags to associate with the image, of the form `{category: value}`.
            """
            if locator in self.data.locators:
                raise ValueError(f"duplicate locator: {locator}")
            self.data.locators.add(locator)
            self.data.records.append(
                (
                    locator,
                    data_source,
                    width,
                    height,
                    None,
                    None,
                    bounding_box,
                    landmarks,
                    tags or {},
                ),
            )

        @validate_arguments(config=ValidatorConfig)
        def add_augmented(
            self,
            original_locator: str,
            augmented_locator: str,
            augmentation_spec: Dict[str, Any],
            width: Optional[int] = None,  # if absent, original width is used
            height: Optional[int] = None,  # if absent, original height is used
            bounding_box: Optional[np.ndarray] = None,  # if absent, original bbox is used if defined
            landmarks: Optional[np.ndarray] = None,  # if absent, original lmks are used if defined
            tags: Optional[Dict[str, str]] = None,  # note that tags are not propagated forward from the original
        ) -> None:
            """
            Add an augmented version of an existing image to Kolena.

            Note that the original image must already be registered in a previous pass. Tags on the original image are
            not propagated forward to the augmented image.

            :param original_locator: The bucket locator for the original version of this image within the platform.
            :param augmented_locator: The bucket locator for the augmented image being registered.
            :param augmentation_spec: Free-form JSON specification for the augmentation applied to this image.
            :param width: Optionally specify the width of the augmented image. When absent, the width of the
                corresponding original image is used.
            :param height: Optionally specify the height of the augmented image. When absent, the height of the
                corresponding original image is used.
            :param bounding_box: Optionally specify a new bounding box for the augmented image. When absent, any
                bounding box corresponding to the original image is used.
            :param landmarks: Optionally specify a new set of landmarks for the augmented image. When absent, any set of
                landmarks corresponding to the original image is used.
            :param tags: Optionally specify a set of tags to associate with the augmented image.
            """
            if augmented_locator in self.data.locators:
                raise ValueError(f"duplicate locator: {augmented_locator}")
            self.data.locators.add(augmented_locator)
            self.data.records.append(
                (
                    augmented_locator,
                    None,
                    width or -1,
                    height or -1,
                    original_locator,
                    augmentation_spec,
                    bounding_box,
                    landmarks,
                    tags or {},
                ),
            )

    @classmethod
    @contextmanager
    def register(cls) -> Iterator[Registrar]:
        """
        Context-managed interface to register new images with Kolena. Images with locators that already exist in the
        platform will have their metadata updated. All changes are committed when the context is exited.

        :raises RemoteError: The registered images were unable to be successfully committed for any reason.
        """
        log.info("registering test images")
        registrar = TestImages.Registrar.__factory__(TestImages._Registrar(records=[], locators=set()))
        yield registrar

        init_response = init_upload()
        df = pd.DataFrame(registrar.data.records, columns=TEST_IMAGE_COLUMNS)
        df["image_id"] = -1
        df_validated = TestImageDataFrame(validate_df_schema(df, TestImageDataFrameSchema))
        df_serializable = df_validated.as_serializable()

        upload_data_frame(df=df_serializable, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=init_response.uuid)
        request = LoadAPI.WithLoadUUID(uuid=init_response.uuid)
        finalize_res = krequests.put(
            endpoint_path=API.Path.COMPLETE_REGISTER.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(finalize_res)
        log.success("registered test images")

    @classmethod
    @validate_arguments(config=ValidatorConfig)
    def iter(
        cls,
        data_source: Optional[Union[str, TestSuite, TestSuite.Data, TestCase, TestCase.Data]] = None,
        include_augmented: bool = False,
        batch_size: int = 10_000_000,
    ) -> Iterator[TestImageDataFrame]:
        """
        Iterator of DataFrames describing images registered in the Kolena platform.

        :param data_source: Optionally specify the single data source to be retrieved, e.g. `"my-data-source"`.
            Alternatively, provide a [`TestSuite`][kolena.fr.TestSuite] or [`TestCase`][kolena.fr.TestCase] as source.
            If no argument is provided, all images registered using
            [`TestImages.register`][kolena.fr.TestImages.register] are returned.
        :param include_augmented: Optionally specify that augmented images should be returned. By default, only
            original images are returned. Ignored when test case or test suite is provided as `data_source`.
        :param batch_size: Optionally specify maximum number of rows to be returned in a single DataFrame.
        """
        test_suite_data = data_source.data if isinstance(data_source, TestSuite) else data_source
        test_suite_id = test_suite_data.id if isinstance(test_suite_data, TestSuite.Data) else None
        test_case_data = data_source.data if isinstance(data_source, TestCase) else data_source
        test_case_id = test_case_data.id if isinstance(test_case_data, TestCase.Data) else None
        data_source_display_name = cls._data_source_display_name(data_source, include_augmented)
        from_extra = f" from '{data_source_display_name}'" if data_source_display_name is not None else ""
        log.info(f"loading test images{from_extra}")
        init_request = API.InitLoadRequest(
            include_augmented=include_augmented,
            data_source=data_source if isinstance(data_source, str) else None,
            test_suite_id=test_suite_id,
            test_case_id=test_case_id,
            batch_size=batch_size,
        )
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_REQUEST.value,
            df_class=TestImageDataFrame,
        )
        log.info(f"loaded test images{from_extra}")

    @staticmethod
    def _data_source_display_name(
        data_source: Optional[Union[str, TestSuite, TestSuite.Data, TestCase, TestCase.Data]],
        include_augmented: bool,
    ) -> Optional[str]:
        if isinstance(data_source, str):
            augmented = " (including augmented images)" if include_augmented else ""
            return f"data source '{data_source}'{augmented}"
        if isinstance(data_source, (TestCase, TestCase.Data)):
            return f"test case '{data_source.data.name if isinstance(data_source, TestCase) else data_source.name}'"
        if isinstance(data_source, (TestSuite, TestSuite.Data)):
            return f"test suite '{data_source.data.name if isinstance(data_source, TestSuite) else data_source.name}'"
        return None



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kolena._api.v1.fr import Asset


class AssetPathMapper:
    """
    Maps images associated with test run as PNG assets to the configured S3 bucket as:
        s3://<bucket>/<database>/<schema>/<test-run-id>/<image-id>/<key>.png
    """

    def __init__(self, config: Asset.Config):
        self.bucket = config.bucket
        self.prefix = config.prefix

    def absolute_locator(self, test_run_id: int, load_uuid: str, image_id: int, key: str) -> str:
        return self._absolute_locator(self.relative_locator(self.path_stub(test_run_id, load_uuid, image_id, key)))

    def relative_locator(self, path_stub: str) -> str:
        return f"{self.prefix}/{path_stub}"

    @staticmethod
    def path_stub(test_run_id: int, load_uuid: str, image_id: int, key: str) -> str:
        return f"{test_run_id}/{image_id}/{key}-{load_uuid}.png"

    def _absolute_locator(self, relative_locator: str) -> str:
        relative_locator_stripped = relative_locator.strip("/")
        return f"s3://{self.bucket}/{relative_locator_stripped}"



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Schema declarations for the [Pandas DataFrames](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html) used
in `kolena.fr`.
"""
import json
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import numpy as np
import pandas as pd
import pandera as pa
from pandera.extensions import register_check_method
from pandera.typing import Series

from kolena._utils.asset_path_mapper import AssetPathMapper
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.datatypes import LoadableDataFrame
from kolena._utils.serde import deserialize_embedding_vector
from kolena._utils.serde import serialize_embedding_vector
from kolena._utils.serde import with_serialized_columns


__ALLOWED_INTEGRAL_DTYPES = {
    int,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int32,
    np.int64,
    np.dtype("uint8"),
    np.dtype("uint16"),
    np.dtype("uint32"),
    np.dtype("uint64"),
    np.dtype("int8"),
    np.dtype("int32"),
    np.dtype("int64"),
}
__ALLOWED_NUMERIC_DTYPES = {
    *__ALLOWED_INTEGRAL_DTYPES,
    float,
    np.float16,
    np.float32,
    np.float64,
    np.longdouble,
    np.dtype("float16"),
    np.dtype("float32"),
    np.dtype("float64"),
    np.dtype("longdouble"),
}


# NOTE: these are actually np.ndarrays, but are declared as objects for compatibility with pandera schemas (as far as
#  pandas is concerned these cells are type 'object')
BoundingBox = object  # bounding box [top_left_x, top_left_y, bottom_right_x, bottom_right_y]
Landmarks = object  # landmarks [left_eye_{x,y}, right_eye_{x,y}, nose_{x,y}, left_mouth_{x,y}, right_mouth_{x,y}]
RGBImage = object  # MxNx3 uint8 ndarray representing an RGB image of width N and height M
EmbeddingVector = object  # embedding vector (arbitrary size and data type)
JSONObject = object


# note that all checks accept nulls -- rely on pandera to properly enforce nullability
def _validate_ndarray(series: Series) -> bool:
    return series.dropna().apply(lambda cell: isinstance(cell, np.ndarray)).all()


@register_check_method()
def _validate_bounding_box(series: Series) -> bool:
    def validate_cell(cell: np.ndarray) -> bool:
        return cell.dtype in __ALLOWED_NUMERIC_DTYPES and cell.shape == (4,)

    return _validate_ndarray(series) and series.dropna().apply(validate_cell).all()


@register_check_method()
def _validate_landmarks(series: Series) -> bool:
    def validate_cell(cell: np.ndarray) -> bool:
        return cell.dtype in __ALLOWED_NUMERIC_DTYPES and cell.shape == (10,)

    return _validate_ndarray(series) and series.dropna().apply(validate_cell).all()


@register_check_method()
def _validate_rgb_image(series: Series) -> bool:
    def validate_cell(cell: np.ndarray) -> bool:
        return cell.dtype == np.uint8 and len(cell.shape) == 3 and cell.shape[2] == 3

    return _validate_ndarray(series) and series.dropna().apply(validate_cell).all()


@register_check_method()
def _validate_embedding_vector(series: Series) -> bool:
    return _validate_ndarray(series)


@register_check_method()
def _validate_optional_dimension(series: Series) -> bool:
    # treat -1 as the value for "not specified"
    return (
        series.dropna().apply(lambda cell: type(cell) in __ALLOWED_INTEGRAL_DTYPES and (cell > 0 or cell == -1)).all()
    )


@register_check_method()
def _validate_json_object(series: Series) -> bool:
    # TODO: more detailed validation? validate that each cell can serialize to JSON
    return series.dropna().apply(lambda cell: isinstance(cell, dict)).all()


@register_check_method()
def _validate_tags(series: Series) -> bool:
    def is_dict_str_str(cell: dict) -> bool:
        return all(isinstance(k, str) and isinstance(v, str) for k, v in cell.items())

    return series.dropna().apply(lambda cell: isinstance(cell, dict) and is_dict_str_str(cell)).all()


def _as_json(value: Optional[Any]) -> Optional[str]:
    return json.dumps(value) if value is not None else None


def _from_json(value: Optional[str]) -> Optional[Any]:
    return json.loads(value) if value is not None else None


class TestImageDataFrameSchema(pa.SchemaModel):
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    """Internal ID corresponding to this image."""

    locator: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    """External locator pointing to image in bucket."""

    data_source: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)
    """Specify source dataset, e.g. "CIFAR-10"."""

    width: Series[pa.typing.Int64] = pa.Field(coerce=True, _validate_optional_dimension=())
    """Width of the image, in pixels. The value `-1` is used to specify "unspecified"."""

    height: Series[pa.typing.Int64] = pa.Field(coerce=True, _validate_optional_dimension=())
    """Height of the image, in pixels. The value `-1` is used to specify "unspecified"."""

    original_locator: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True, _validate_locator=())
    """Specify that this image is an augmented version of another (registered) image."""

    augmentation_spec: Series[JSONObject] = pa.Field(coerce=True, nullable=True, _validate_json_object=())
    """
    Free-form specification describing the augmentation applied to the image, e.g. `{"rotate": 90}`. Should not be
    specified unless original_locator is present.
    """

    bounding_box: Series[BoundingBox] = pa.Field(coerce=True, nullable=True, _validate_bounding_box=())
    """Ground truth bounding box. If absent, no ground truth will be available for display."""

    landmarks: Series[Landmarks] = pa.Field(coerce=True, nullable=True, _validate_landmarks=())
    """Ground truth landmarks. If absent, no ground truth will be available for display."""

    tags: Series[JSONObject] = pa.Field(coerce=True, _validate_tags=())
    """
    Specify a set of tags to apply to this object in the form `{category: value}`. Note that this format
    intentionally restricts tags to a single value per category.
    """


TestImageRecord = Tuple[
    str,
    Optional[str],
    int,
    int,
    Optional[str],
    Optional[Dict[str, Any]],
    Optional[np.ndarray],
    Optional[np.ndarray],
    Dict[str, Any],
]
TEST_IMAGE_COLUMNS = [
    "locator",
    "data_source",
    "width",
    "height",
    "original_locator",
    "augmentation_spec",
    "bounding_box",
    "landmarks",
    "tags",
]


class TestCaseDataFrameSchema(pa.SchemaModel):
    locator_a: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    locator_b: Series[pa.typing.String] = pa.Field(coerce=True, _validate_locator=())
    is_same: Series[pa.typing.Bool] = pa.Field(coerce=True)


TestCaseRecord = Tuple[str, str, bool]
TEST_CASE_COLUMNS = ["locator_a", "locator_b", "is_same"]


class ImageDataFrameSchema(pa.SchemaModel):
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    locator: Series[pa.typing.String] = pa.Field(coerce=True)


class ImageResultDataFrameSchema(pa.SchemaModel):
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    """[**Required**] The ID of the image corresponding to this record."""

    bounding_box: Optional[Series[BoundingBox]] = pa.Field(coerce=True, nullable=True, _validate_bounding_box=())
    """[Optional] A bounding box around the face detected in this image."""

    landmarks_input_image: Optional[Series[RGBImage]] = pa.Field(coerce=True, nullable=True, _validate_rgb_image=())
    """
    [Optional] RGB image (`np.ndarray` with cells of type `np.uint8`) corresponding to the input to the "landmarks"
    model in the face recognition pipeline.
    """

    landmarks: Optional[Series[Landmarks]] = pa.Field(coerce=True, nullable=True, _validate_landmarks=())
    """
    [Optional] A 10-element array with `(x, y)` coordinates corresponding to the left eye, right eye, nose tip, left
    mouth corner, right mouth corner of the detected face.
    """

    quality_input_image: Optional[Series[RGBImage]] = pa.Field(coerce=True, nullable=True, _validate_rgb_image=())
    """
    [Optional] RGB image (`np.ndarray` with cells of type `np.uint8`) corresponding to the input to the "quality" model
    in the face recognition pipeline.
    """

    quality: Optional[Series[pa.typing.Float64]] = pa.Field(coerce=True, nullable=True)
    """[Optional] Score produced by the "quality" model in the face recognition pipeline."""

    acceptability: Optional[Series[pa.typing.Float64]] = pa.Field(coerce=True, nullable=True)
    """[Optional] Score produced by the "acceptability" model in the face recognition pipeline."""

    fr_input_image: Optional[Series[RGBImage]] = pa.Field(coerce=True, nullable=True, _validate_rgb_image=())
    """
    [Optional] RGB image (`np.ndarray` with cells of type `np.uint8`) corresponding to the input to the facial
    embeddings extraction model in the face recognition pipeline.
    """

    embedding: Series[EmbeddingVector] = pa.Field(coerce=True, nullable=True, _validate_embedding_vector=())
    """
    [Optional] Embedding vector (`np.ndarray`) extracted representing the face detected in the image. An empty cell
    (no array is provided) indicates a failure to enroll for the image, i.e. no face detected.
    """

    failure_reason: Optional[Series[pa.typing.String]] = pa.Field(coerce=True, nullable=True)
    """[Optional] The reason why the image was a failure to enroll."""


class _ResultStageFrameSchema(pa.SchemaModel):
    test_run_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    # null values for the following columns indicate a failure to enroll
    bbox: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)
    lmks: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)
    embedding: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)
    assets: Series[pa.typing.String] = pa.Field(coerce=True)  # should be {} when no data is present
    metadata: Series[pa.typing.String] = pa.Field(coerce=True)  # should be {} when no data is present
    failure_reason: Series[pa.typing.String] = pa.Field(coerce=True, nullable=True)  # should be null unless failure


class _ImageChipsDataFrameSchema(pa.SchemaModel):
    test_run_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    key: Series[pa.typing.String] = pa.Field(coerce=True)
    uuid: Series[pa.typing.String] = pa.Field(coerce=True)
    image: Series[RGBImage] = pa.Field(coerce=True, _validate_rgb_image=())


class EmbeddingDataFrameSchema(pa.SchemaModel):
    image_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    """The provided ID of the image."""

    embedding: Series[EmbeddingVector] = pa.Field(nullable=True, coerce=True, _validate_embedding_vector=())
    """
    The extracted embedding(s) corresponding to the `image_id`. A missing embedding indicates a failure to enroll
    for the image.

    For images with only one extracted embedding, the `embedding` is a one-dimensional `np.ndarray` with length
    matching the length of the extracted embedding. When multiple embeddings were extracted from a single image, the
    first dimension represents the index of the extracted embedding. For example, for an image with 3 extracted
    embeddings and embeddings of length 256, the `embedding` is an array of shape (3, 256).
    """


class PairDataFrameSchema(pa.SchemaModel):
    image_pair_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    image_a_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    image_b_id: Series[pa.typing.Int64] = pa.Field(coerce=True)


class PairResultDataFrameSchema(pa.SchemaModel):
    image_pair_id: Series[pa.typing.Int64] = pa.Field(coerce=True)
    """[**Required**] The ID of the image corresponding to this record."""

    similarity: Series[pa.typing.Float64] = pa.Field(nullable=True, coerce=True)
    """
    [Optional] The similarity score computed between the two embeddings in this image pair. Should be left empty when
    either image in the pair is a failure to enroll.
    """

    embedding_a_index: Optional[Series[pa.typing.Int64]] = pa.Field(nullable=True, coerce=True)
    """
    [Optional] Index of the embedding in `image_a` corresponding to this similarity score. Required when multiple
    embeddings are extracted per image and multiple similarity scores are computed per image pair.
    """

    embedding_b_index: Optional[Series[pa.typing.Int64]] = pa.Field(nullable=True, coerce=True)
    """
    [Optional] Index of the embedding in `image_b` corresponding to this similarity score. Required when multiple
    embeddings are extracted per image and multiple similarity scores are computed per image pair.
    """


class LoadedPairResultDataFrameSchema(TestCaseDataFrameSchema, PairResultDataFrameSchema):  # note inheritance
    image_a_fte: Series[pa.typing.Bool] = pa.Field(coerce=True)  # if image_a failed to enroll (FTE)
    image_b_fte: Series[pa.typing.Bool] = pa.Field(coerce=True)  # if image_b failed to enroll (FTE)


class TestImageDataFrame(pa.typing.DataFrame[TestImageDataFrameSchema], LoadableDataFrame["TestImageDataFrame"]):
    def as_serializable(self) -> pd.DataFrame:
        object_columns = ["augmentation_spec", "bounding_box", "landmarks", "tags"]
        return with_serialized_columns(self, object_columns)

    @classmethod
    def get_schema(cls) -> Type[TestImageDataFrameSchema]:
        return TestImageDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "TestImageDataFrame":
        object_columns = ["augmentation_spec", "bounding_box", "landmarks", "tags"]
        for col in object_columns:
            df[col] = df[col].apply(_from_json)
        ndarray_columns = ["bounding_box", "landmarks"]
        for col in ndarray_columns:
            df[col] = df[col].apply(lambda v: None if v is None else np.array(v))
        return cast(TestImageDataFrame, validate_df_schema(df, TestImageDataFrameSchema, trusted=True))


class TestCaseDataFrame(pa.typing.DataFrame[TestCaseDataFrameSchema], LoadableDataFrame["TestCaseDataFrame"]):
    @classmethod
    def get_schema(cls) -> Type[TestCaseDataFrameSchema]:
        return TestCaseDataFrameSchema


class LoadedPairResultDataFrame(
    pa.typing.DataFrame[LoadedPairResultDataFrameSchema],
    LoadableDataFrame["LoadedPairResultDataFrame"],
):
    @classmethod
    def get_schema(cls) -> Type[LoadedPairResultDataFrameSchema]:
        return LoadedPairResultDataFrameSchema


class ImageResultDataFrame(pa.typing.DataFrame[ImageResultDataFrameSchema]):
    ...


class _ResultStageFrame(pa.typing.DataFrame[_ResultStageFrameSchema]):
    @classmethod
    def from_image_result_data_frame(
        cls,
        test_run_id: int,
        load_uuid: str,
        df: ImageResultDataFrame,
        path_mapper: AssetPathMapper,
    ) -> "_ResultStageFrame":
        df_stage = pd.DataFrame(
            dict(
                test_run_id=test_run_id,
                image_id=df["image_id"],
                bbox=[
                    json.dumps(record.bounding_box.tolist())
                    if "bounding_box" in df.columns and record.bounding_box is not None
                    else None
                    for record in df.itertuples()
                ],
                lmks=[
                    json.dumps(record.landmarks.tolist())
                    if "landmarks" in df.columns and record.landmarks is not None
                    else None
                    for record in df.itertuples()
                ],
                embedding=[
                    serialize_embedding_vector(record.embedding) if record.embedding is not None else None
                    for record in df.itertuples()
                ],
                assets=[
                    json.dumps(
                        {
                            f"{key}_input_image": path_mapper.absolute_locator(
                                test_run_id=test_run_id,
                                load_uuid=load_uuid,
                                image_id=getattr(record, "image_id"),
                                key=key,
                            )
                            for key in ["landmarks", "quality", "fr"]
                            if f"{key}_input_image" in df.columns and getattr(record, f"{key}_input_image") is not None
                        },
                    )
                    for record in df.itertuples()
                ],
                metadata=[
                    json.dumps(
                        {
                            key: getattr(record, key)
                            for key in ["quality", "acceptability"]
                            # this works because both quality and acceptability are float types
                            if key in df.columns and not pd.isna(getattr(record, key))
                        },
                    )
                    for record in df.itertuples()
                ],
                failure_reason=df["failure_reason"] if "failure_reason" in df.columns else [None] * len(df),
            ),
        )
        return cast(_ResultStageFrame, validate_df_schema(df_stage, _ResultStageFrameSchema, trusted=True))


class _ImageChipsDataFrame(pa.typing.DataFrame[_ImageChipsDataFrameSchema]):
    @classmethod
    def from_image_result_data_frame(
        cls,
        test_run_id: int,
        load_uuid: str,
        df: ImageResultDataFrame,
    ) -> "_ImageChipsDataFrame":
        keys = ["landmarks", "quality", "fr"]
        df_chips_locators: List[pd.DataFrame] = []
        for key in keys:
            image_col = f"{key}_input_image"
            if image_col not in df.columns:
                columns = ["image_id", "image", "test_run_id", "uuid", "key"]
                df_chips_locators.append(pd.DataFrame([], columns=columns))
                continue
            df_key = df[["image_id", image_col]]
            df_key = df_key[df_key[image_col].notnull()]
            df_key["test_run_id"] = test_run_id
            df_key["uuid"] = load_uuid
            df_key["key"] = key
            df_key = df_key.rename(columns={image_col: "image"})
            df_chips_locators.append(df_key)
        df_merged = pd.concat(df_chips_locators)
        return cls(validate_df_schema(df_merged, _ImageChipsDataFrameSchema, trusted=True))


class ImageDataFrame(pa.typing.DataFrame[ImageDataFrameSchema], LoadableDataFrame["ImageDataFrame"]):
    @classmethod
    def get_schema(cls) -> Type[ImageDataFrameSchema]:
        return ImageDataFrameSchema


class EmbeddingDataFrame(pa.typing.DataFrame[EmbeddingDataFrameSchema], LoadableDataFrame["EmbeddingDataFrame"]):
    @classmethod
    def get_schema(cls) -> Type[EmbeddingDataFrameSchema]:
        return EmbeddingDataFrameSchema

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> "EmbeddingDataFrame":
        records: List[Tuple[int, Optional[np.ndarray]]] = []
        for image_id, df_embedding in df.groupby("image_id"):
            image_id_int = int(image_id)  # type: ignore
            embeddings = [  # prevented from including both embeddings and None (FTE) during ingest
                deserialize_embedding_vector(record.embedding)
                for record in df_embedding.itertuples()
                if record.embedding is not None
            ]
            if len(embeddings) == 0:
                records.append((image_id_int, None))
            elif len(embeddings) == 1:
                records.append((image_id_int, embeddings[0]))
            else:
                embeddings_arr = np.concatenate([embedding.reshape(1, len(embedding)) for embedding in embeddings])
                records.append((image_id_int, embeddings_arr))
        df_stage = pd.DataFrame(records, columns=["image_id", "embedding"])
        return cast(EmbeddingDataFrame, validate_df_schema(df_stage, EmbeddingDataFrameSchema, trusted=True))


class PairDataFrame(pa.typing.DataFrame[PairDataFrameSchema], LoadableDataFrame["PairDataFrame"]):
    @classmethod
    def get_schema(cls) -> Type[PairDataFrameSchema]:
        return PairDataFrameSchema


PairResultDataFrame = pa.typing.DataFrame[PairResultDataFrameSchema]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Union

import numpy as np
from pydantic.dataclasses import dataclass

from kolena._api.v1.fr import Model as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.endpoints import get_model_url
from kolena._utils.serde import from_dict
from kolena._utils.uninstantiable import Uninstantiable
from kolena._utils.validators import ValidatorConfig
from kolena.fr import TestCase
from kolena.fr import TestSuite
from kolena.fr.datatypes import LoadedPairResultDataFrame


class Model(Uninstantiable["Model.Data"]):
    """The descriptor for your model within the Kolena platform."""

    @dataclass(frozen=True, config=ValidatorConfig)
    class Data:
        id: int
        name: str
        metadata: Dict[str, Any]

    @classmethod
    def create(cls, name: str, metadata: Dict[str, Any]) -> "Model":
        """
        Create a new model with the provided name and metadata.

        :param name: Unique name of the new model to create.
        :param metadata: Unstructured metadata to associate with the model.
        :return: The newly created model.
        :raises ValueError: A model by the provided name already exists.
        """
        request = API.CreateRequest(name=name, metadata=metadata)
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = Model.__factory__(from_dict(data_class=Model.Data, data=res.json()))
        log.info(f"created model '{name}' ({get_model_url(obj.data.id)})")
        return obj

    @classmethod
    def load_by_name(cls, name: str) -> "Model":
        """
        Retrieve the existing model with the provided name.

        :param name: Name of the model to retrieve.
        :return: The retrieved model.
        :raises KeyError: If no model with the provided name exists.
        """
        request = API.LoadByNameRequest(name=name)
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        obj = Model.__factory__(from_dict(data_class=Model.Data, data=res.json()))
        log.info(f"loaded model '{name}' ({get_model_url(obj.data.id)})")
        return obj

    def _get_load_pair_results_request(
        self,
        test_object: Union[TestSuite, TestSuite.Data, TestCase, TestCase.Data],
    ) -> API.LoadPairResultsRequest:
        if isinstance(test_object, TestSuite):
            return API.LoadPairResultsRequest(model_id=self.data.id, test_suite_id=test_object.data.id)
        elif isinstance(test_object, TestSuite.Data):
            return API.LoadPairResultsRequest(model_id=self.data.id, test_suite_id=test_object.id)
        elif isinstance(test_object, TestCase):
            return API.LoadPairResultsRequest(model_id=self.data.id, test_case_id=test_object.data.id)
        elif isinstance(test_object, TestCase.Data):
            return API.LoadPairResultsRequest(model_id=self.data.id, test_case_id=test_object.id)
        else:
            raise ValueError(f"invalid test object type provided: {type(test_object).__name__}")

    def load_pair_results(
        self,
        test_object: Union[TestSuite, TestSuite.Data, TestCase, TestCase.Data],
    ) -> LoadedPairResultDataFrame:
        """
        Load previously stored pair results for this model on the provided test case or test suite. If this model has
        not been run on the provided test object, a zero-length response is returned. Partial results are returned when
        testing on the requested test case or test suite is incomplete.

        The returned DataFrame has the following relevant fields:

        - `locator_a`: the locator pointing to the left image in the pair
        - `locator_b`: the locator pointing to the right image in the pair
        - `is_same`: boolean indicating if the two images depict the same person or a different person
        - `image_a_fte`: boolean indicating that the left image failed to enroll (FTE)
        - `image_b_fte`: boolean indicating that the right image failed to enroll (FTE)
        - `similarity`: float similarity score between the left and right images. `NaN` if either image failed to
          enroll. When multiple similarity scores were provided for a given image pair, only the highest similarity
          score is returned

        :param test_object: The [`TestSuite`][kolena.fr.TestSuite] or [`TestCase`][kolena.fr.TestCase] to load pair
            results from.
        :raises ValueError: An invalid test object was provided.
        :raises RemoteError: The pair results could not be loaded for any reason.
        """
        return _BatchedLoader.concat(self.iter_pair_results(test_object), LoadedPairResultDataFrame)

    def iter_pair_results(
        self,
        test_object: Union[TestSuite, TestSuite.Data, TestCase, TestCase.Data],
        batch_size: int = 10_000_000,
    ) -> Iterator[LoadedPairResultDataFrame]:
        """
        Iterator over DataFrames of previously stored pair results for this model on the provided test case or test
        suite, grouped in batches. If this model has not been run on the provided test object, a zero-length response
        is returned. Partial results are returned when testing on the requested test case or test suite is incomplete.

        See [`Model.load_pair_results`][kolena.fr.Model.load_pair_results] for details on the returned DataFrame.

        :param test_object: The [`TestSuite`][kolena.fr.TestSuite] or [`TestCase`][kolena.fr.TestCase] to load pair
            results from.
        :param batch_size: Optionally specify maximum number of rows to be returned in a single DataFrame.
        :raises ValueError: An invalid test object was provided.
        :raises RemoteError: The pair results could not be loaded for any reason.
        """
        display_type = "test case" if isinstance(test_object, (TestCase, TestCase.Data)) else "test suite"
        display_name = test_object.data.name if isinstance(test_object, (TestCase, TestSuite)) else test_object.name
        test_object_display_name = f"{display_type} '{display_name}'"
        log.info(f"loading pair results from model '{self.data.name}' on {test_object_display_name}")
        base_load_request = dataclasses.asdict(self._get_load_pair_results_request(test_object))
        init_request = API.InitLoadPairResultsRequest(batch_size=batch_size, **base_load_request)
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_PAIR_RESULTS.value,
            df_class=LoadedPairResultDataFrame,
        )
        log.info(f"loaded pair results from model '{self.data.name}' on {test_object_display_name}")


class InferenceModel(Model):
    """
    A [`Model`][kolena.fr.Model] capable of running tests via [`test`][kolena.fr.test].

    Currently supports extracting a single embedding per image. To extract multiple embeddings per image, see
    [`TestRun`][kolena.fr.TestRun].
    """

    extract: Callable[[str], Optional[np.ndarray]]
    compare: Callable[[np.ndarray, np.ndarray], float]

    @classmethod
    def create(
        cls,
        name: str,
        extract: Callable[[str], Optional[np.ndarray]],
        compare: Callable[[np.ndarray, np.ndarray], float],
        metadata: Dict[str, Any],
    ) -> "InferenceModel":
        """
        Create a new model with the provided name and metadata.

        :param name: Unique name of the new model to create.
        :param extract: A function implementing embeddings extraction for this model.
        :param compare: A function implementing embeddings similarity comparison for this model.
        :param metadata: Unstructured metadata to associate with the model.
        :return: The newly created model.
        :raises ValueError: A model by the provided name already exists.
        """
        base_model = super().create(name, metadata)
        return cls._from_base(base_model, extract, compare)

    @classmethod
    def load_by_name(
        cls,
        name: str,
        extract: Callable[[str], Optional[np.ndarray]],
        compare: Callable[[np.ndarray, np.ndarray], float],
    ) -> "InferenceModel":
        """
        Load an existing model.

        :param name: The name of the model to load.
        :param extract: A function implementing embeddings extraction for this model.
        :param compare: A function implementing embeddings similarity comparison for this model.
        :return: The loaded model.
        """
        base_model = super().load_by_name(name)
        return cls._from_base(base_model, extract, compare)

    @staticmethod
    def _from_base(
        base_model: Model,
        extract: Callable[[str], Optional[np.ndarray]],
        compare: Callable[[np.ndarray, np.ndarray], float],
    ) -> "InferenceModel":
        inference_model = InferenceModel.__factory__(base_model.data)
        object.__setattr__(inference_model, "__frozen__", False)
        setattr(inference_model, "extract", extract)  # bypass mypy method assignment bug
        setattr(inference_model, "compare", compare)  # see: https://github.com/python/mypy/issues/2427
        inference_model.__frozen__ = True
        return inference_model



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from abc import ABCMeta
from abc import abstractmethod
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from pydantic import validate_arguments

from kolena._utils.frozen import Frozen
from kolena._utils.serializable import Serializable
from kolena._utils.validators import ValidatorConfig
from kolena.errors import InputValidationError


class MetadataElementType(str, Enum):
    BOUNDING_BOX = "BOUNDING_BOX"
    LANDMARKS = "LANDMARKS"  # TODO: migrate terminology to "keypoints", including existing db data.
    ASSET = "ASSET"


class Annotation(Frozen, Serializable, metaclass=ABCMeta):
    """
    An annotation associated with an image.

    Annotations are surfaced during testing along with the image locator and any other metadata associated with an
    image. In the web platform, annotations are overlaid on top of images when visualizing results.
    """

    @abstractmethod
    def __init__(self) -> None:
        ...


class Asset(Frozen, Serializable):
    """
    An asset living in your shared bucket. Assets are surfaced during testing along with any other metadata associated
    with a given test image.

    In the web platform, certain assets such as PNG and JPG images are viewable when visualizing results in the gallery.
    """

    locator: str
    """Location of this asset in shared bucket, e.g. `s3://my-bucket/path/to/image.png`."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, locator: str):
        self.locator = locator  # TODO: validate locator

    def _to_dict(self) -> Dict[str, Any]:
        return dict(data_type=MetadataElementType.ASSET.value, data_object=dict(locator=self.locator))


MetadataScalar = Union[None, str, float, int, bool]
MetadataElement = Union[MetadataScalar, Annotation, Asset]


def _to_dict(metadata: Dict[str, MetadataElement]) -> Dict[str, Any]:
    def dump_element(element: MetadataElement) -> Any:
        if isinstance(element, Serializable):
            return element._to_dict()
        if isinstance(element, float):
            if math.isnan(element):
                return None
            if math.isinf(element):
                raise InputValidationError("cannot serialize infinite metadata values")
        return element

    return {key: dump_element(value) for key, value in metadata.items()}


class BoundingBox(Annotation):
    """An annotation comprising a bounding box around an object in an image."""

    top_left: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the top left corner of the bounding box."""

    bottom_right: Tuple[float, float]
    """Point in `(x, y)` pixel coordinates representing the bottom right corner of the bounding box."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, top_left: Tuple[float, float], bottom_right: Tuple[float, float]):
        self.top_left = top_left
        self.bottom_right = bottom_right

    def _to_dict(self) -> Dict[str, Any]:
        return dict(
            data_type=MetadataElementType.BOUNDING_BOX.value,
            data_object=dict(points=[self.top_left, self.bottom_right]),
        )


class Landmarks(Annotation):
    """
    An annotation comprising an arbitrary-length set of landmarks corresponding to some object in an image, e.g. face
    landmarks used for pose estimation.
    """

    points: List[Tuple[float, float]]
    """Any number of `(x, y)` points in pixel coordinates representing a set of landmarks."""

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, points: List[Tuple[float, float]]):
        if len(points) == 0:
            raise ValueError("At least one point required for landmarks annotation.")
        self.points = points

    def _to_dict(self) -> Dict[str, Any]:
        return dict(data_type=MetadataElementType.LANDMARKS.value, data_object=dict(points=self.points))


def _parse_element(element: Union[None, str, float, int, bool, Dict[str, Any]]) -> MetadataElement:
    if element is None or isinstance(element, (str, float, int, bool)):
        return element
    if isinstance(element, dict):
        data_type = element["data_type"]
        data_object = element["data_object"]
        if data_type == MetadataElementType.BOUNDING_BOX.value:
            top_left, bottom_right = data_object["points"]
            return BoundingBox(top_left, bottom_right)
        if data_type == MetadataElementType.LANDMARKS.value:
            return Landmarks(data_object["points"])
        if data_type == MetadataElementType.ASSET.value:
            return Asset(data_object["locator"])
    raise ValueError(f"unrecognized value: {element}")


def _from_dict(metadata_blob: Dict[str, Any]) -> Dict[str, MetadataElement]:
    return {key: _parse_element(value) for key, value in metadata_blob.items()}



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from enum import Enum
from typing import List
from typing import Optional

from pydantic.dataclasses import dataclass

from kolena.workflow import define_workflow
from kolena.workflow import EvaluatorConfiguration
from kolena.workflow import GroundTruth as BaseGroundTruth
from kolena.workflow import Image
from kolena.workflow import Inference as BaseInference
from kolena.workflow import Metadata
from kolena.workflow import MetricsTestCase
from kolena.workflow import MetricsTestSample
from kolena.workflow import MetricsTestSuite
from kolena.workflow.annotation import LabeledBoundingBox
from kolena.workflow.annotation import ScoredClassificationLabel
from kolena.workflow.annotation import ScoredLabeledBoundingBox


@dataclass(frozen=True)
class TestSample(Image):
    metadata: Metadata = dataclasses.field(default_factory=dict)


@dataclass(frozen=True)
class GroundTruth(BaseGroundTruth):
    bboxes: List[LabeledBoundingBox]
    ignored_bboxes: List[LabeledBoundingBox] = dataclasses.field(default_factory=list)
    labels: List[str] = dataclasses.field(default_factory=list)
    n_bboxes: int = dataclasses.field(default_factory=lambda: 0)

    def __post_init__(self):
        object.__setattr__(self, "labels", sorted({box.label for box in self.bboxes}))
        object.__setattr__(self, "n_bboxes", len(self.bboxes))


@dataclass(frozen=True)
class Inference(BaseInference):
    bboxes: List[ScoredLabeledBoundingBox]


_workflow, TestCase, TestSuite, Model = define_workflow(
    "Object Detection",
    TestSample,
    GroundTruth,
    Inference,
)


@dataclass(frozen=True)
class TestSampleMetrics(MetricsTestSample):
    TP: List[ScoredLabeledBoundingBox]
    FP: List[ScoredLabeledBoundingBox]
    FN: List[LabeledBoundingBox]
    Confused: List[ScoredLabeledBoundingBox]

    count_TP: int
    count_FP: int
    count_FN: int
    count_Confused: int

    has_TP: bool
    has_FP: bool
    has_FN: bool
    has_Confused: bool

    max_confidence_above_t: Optional[float]
    min_confidence_above_t: Optional[float]
    thresholds: List[ScoredClassificationLabel]


@dataclass(frozen=True)
class ClassMetricsPerTestCase(MetricsTestCase):
    Class: str
    nImages: int
    Threshold: float
    Objects: int
    Inferences: int
    TP: int
    FN: int
    FP: int
    Precision: float
    Recall: float
    F1: float
    AP: float


@dataclass(frozen=True)
class TestCaseMetrics(MetricsTestCase):
    PerClass: List[ClassMetricsPerTestCase]
    Objects: int
    Inferences: int
    TP: int
    FN: int
    FP: int
    macro_Precision: float
    macro_Recall: float
    macro_F1: float
    mean_AP: float


@dataclass(frozen=True)
class TestSuiteMetrics(MetricsTestSuite):
    n_images: int
    mean_AP: float


class ThresholdStrategy(str, Enum):
    F1_OPTIMAL = "F1_OPTIMAL"
    FIXED_03 = "FIXED_03"
    FIXED_05 = "FIXED_05"
    FIXED_075 = "FIXED_075"

    def display_name(self) -> str:
        if self is ThresholdStrategy.FIXED_03:
            return "Fixed(0.3)"
        if self is ThresholdStrategy.FIXED_05:
            return "Fixed(0.5)"
        if self is ThresholdStrategy.FIXED_075:
            return "Fixed(0.75)"
        if self is ThresholdStrategy.F1_OPTIMAL:
            return "F1-Optimal"
        raise RuntimeError(f"unrecognized threshold strategy: {self}")


@dataclass(frozen=True)
class ThresholdConfiguration(EvaluatorConfiguration):
    threshold_strategy: ThresholdStrategy
    iou_threshold: float
    min_confidence_score: float

    def display_name(self) -> str:
        return (
            f"Threshold: {self.threshold_strategy.display_name()}, "
            f"IoU: {self.iou_threshold}, confidence ≥ {self.min_confidence_score}"
        )



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from typing import Any
from typing import Dict
from typing import Optional

import pandas as pd

from kolena._utils import log
from kolena._utils import state
from kolena._utils.endpoints import get_platform_url
from kolena._utils.instrumentation import upload_log
from kolena._utils.state import _client_state
from kolena.errors import InputValidationError


def initialize(
    api_token: str,
    *args: Any,
    verbose: bool = False,
    proxies: Optional[Dict[str, str]] = None,
    **kwargs: Any,
) -> None:
    """
    Initialize a client session.

    Retrieve an API token from the [:kolena-developer-16: Developer](https://app.kolena.io/redirect/developer) page and
    populate the `KOLENA_TOKEN` environment variable before initializing:

    ```python
    import os
    import kolena

    kolena.initialize(os.environ["KOLENA_TOKEN"], verbose=True)
    ```

    A session has a global scope and remains active until interpreter shutdown.

    !!! note
        As of version 0.29.0: the `entity` argument is no longer needed; the signature `initialize(entity, api_token)`
        has been deprecated and replaced by `initialize(api_token)`.

    :param api_token: Provided API token. This token is a secret and should be treated with caution.
    :param verbose: Optionally configure client to run in verbose mode, providing more information about execution. All
        logging events are emitted as Python standard library `logging` events from the `"kolena"` logger as well as
        to stdout/stderr directly.
    :param proxies: Optionally configure client to run with `http` or `https` proxies. The `proxies` parameter
        is passed through to the `requests` package and can be
        [configured accordingly](https://requests.readthedocs.io/en/latest/user/advanced/#proxies).
    :raises InvalidTokenError: The provided `api_token` is not valid.
    :raises InputValidationError: The provided combination or number of args is not valid.
    """
    used_deprecated_signature = False

    if len(args) > 1:
        raise InputValidationError(f"Too many args. Expected 0 or 1 but got {len(args)} Check docs for usage.")
    elif len(args) == 1:
        # overwrite the originally passed api_token since we are supporting backward compatability with entity
        api_token = args[0]
    if len(args) == 1 or "entity" in kwargs:
        used_deprecated_signature = True
        warnings.warn(
            "The signature initialize(entity, token) is deprecated. Please update to initialize(token).",
            category=DeprecationWarning,
            stacklevel=2,
        )

    init_response = state.get_token(api_token, proxies=proxies)
    derived_telemetry = init_response.tenant_telemetry
    _client_state.update(
        api_token=api_token,
        jwt_token=init_response.access_token,
        tenant=init_response.tenant,
        verbose=verbose,
        telemetry=derived_telemetry,
        proxies=proxies,
    )

    if used_deprecated_signature:
        upload_log("Client attempted to use deprecated entity auth signature.", "warn")

    log.info("initialized")
    if verbose:
        # Configure third party logging based on verbosity
        pd.set_option("display.max_colwidth", None)
        log.info(f"connected to {get_platform_url()}")
